@@ -1,6 +1,6 @@
 # This file is part of Xpra.
 # Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2011-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2011-2022 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -8,9 +8,11 @@ import os
 import sys
 import struct
 
-from xpra.os_util import bytestostr, hexstr
-from xpra.util import iround, envbool, envint, csv, ellipsizer
-from xpra.os_util import is_unity, is_gnome, is_kde, is_Ubuntu, is_Fedora, is_X11, is_Wayland, saved_env
+from xpra.os_util import (
+    bytestostr, hexstr, get_saved_env,
+    is_X11, is_Wayland, get_saved_env_var,
+    )
+from xpra.util import u, envbool, envint, csv, ellipsizer
 from xpra.log import Logger
 
 log = Logger("posix")
@@ -65,7 +67,7 @@ device_bell = None
 GTK_MENUS = envbool("XPRA_GTK_MENUS", False)
 RANDR_DPI = envbool("XPRA_RANDR_DPI", True)
 XSETTINGS_DPI = envbool("XPRA_XSETTINGS_DPI", True)
-USE_NATIVE_TRAY = envbool("XPRA_USE_NATIVE_TRAY", is_unity() or (is_Ubuntu() and is_gnome()) or (is_gnome() and not is_Fedora()) or is_kde())
+USE_NATIVE_TRAY = envbool("XPRA_USE_NATIVE_TRAY", True)
 XINPUT_WHEEL_DIV = envint("XPRA_XINPUT_WHEEL_DIV", 15)
 DBUS_SCREENSAVER = envbool("XPRA_DBUS_SCREENSAVER", False)
 
@@ -79,29 +81,40 @@ def gl_check():
 def get_native_system_tray_classes():
     return []
 
+
 def get_wm_name():
-    wm_name = os.environ.get("XDG_CURRENT_DESKTOP", "") or os.environ.get("XDG_SESSION_DESKTOP") or os.environ.get("DESKTOP_SESSION")
-    if os.environ.get("XDG_SESSION_TYPE")=="wayland" or os.environ.get("GDK_BACKEND")=="wayland":
+    return do_get_wm_name(get_saved_env())
+
+def do_get_wm_name(env):
+    wm_name = env.get("XDG_CURRENT_DESKTOP", "") or env.get("XDG_SESSION_DESKTOP") or env.get("DESKTOP_SESSION")
+    if env.get("XDG_SESSION_TYPE")=="wayland" or env.get("GDK_BACKEND")=="wayland":
         if wm_name:
             wm_name += " on wayland"
         else:
             wm_name = "wayland"
     elif is_X11():
-        try:
-            wm_check = _get_X11_root_property("_NET_SUPPORTING_WM_CHECK", "WINDOW")
-            if wm_check:
-                xid = struct.unpack(b"@L", wm_check)[0]
-                traylog("_NET_SUPPORTING_WM_CHECK window=%#x", xid)
-                wm_name = _get_X11_window_property(xid, "_NET_WM_NAME", "UTF8_STRING")
-                traylog("_NET_WM_NAME=%s", wm_name)
-                if wm_name:
-                    return wm_name.decode("utf-8")
-        except Exception as e:
-            traylog("get_wm_name()", exc_info=True)
-            traylog.error("Error accessing window manager information:")
-            traylog.error(" %s", e)
+        wm_name = get_x11_wm_name()
     return wm_name
 
+def get_x11_wm_name():
+    if not is_X11():
+        return None
+    try:
+        wm_check = _get_X11_root_property("_NET_SUPPORTING_WM_CHECK", "WINDOW")
+        if wm_check:
+            xid = struct.unpack(b"@L", wm_check)[0]
+            traylog("_NET_SUPPORTING_WM_CHECK window=%#x", xid)
+            wm_name = _get_X11_window_property(xid, "_NET_WM_NAME", "UTF8_STRING")
+            traylog("_NET_WM_NAME=%s", wm_name)
+            if wm_name:
+                return u(wm_name)
+    except Exception as e:
+        screenlog("get_x11_wm_name()", exc_info=True)
+        screenlog.error("Error accessing window manager information:")
+        screenlog.error(" %s", e)
+    return None
+
+
 def get_clipboard_native_class():
     if is_Wayland():
         return "xpra.gtk_common.gtk_clipboard.GTK_Clipboard"
@@ -119,7 +132,7 @@ def get_native_tray_classes():
             traylog.warn("Warning: appindicator library not found")
             traylog.warn(" you may want to install libappindicator")
             traylog.warn(" to enable the system tray.")
-            if saved_env.get("XDG_CURRENT_DESKTOP")=="GNOME":
+            if get_saved_env_var("XDG_CURRENT_DESKTOP", "").upper().find("GNOME")>=0:
                 traylog.warn(" With gnome-shell, you may also need some extensions:")
                 traylog.warn(" 'top icons plus' and / or 'appindicator'")
     traylog("get_native_tray_classes()=%s (USE_NATIVE_TRAY=%s)", c, USE_NATIVE_TRAY)
@@ -188,7 +201,7 @@ def _get_xsettings():
         if not data:
             return None
         from xpra.x11.xsettings_prop import get_settings
-        return get_settings(X11Window.get_display_name(), data)
+        return get_settings(data)
     return None
 
 def _get_xsettings_dict():
@@ -230,8 +243,8 @@ def _get_randr_dpi():
                 wmm, hmm = randr_bindings.get_screen_size_mm()
                 if wmm>0 and hmm>0:
                     w, h =  randr_bindings.get_screen_size()
-                    dpix = iround(w * 25.4 / wmm)
-                    dpiy = iround(h * 25.4 / hmm)
+                    dpix = round(w * 25.4 / wmm)
+                    dpiy = round(h * 25.4 / hmm)
                     screenlog("xdpi=%s, ydpi=%s - size-mm=%ix%i, size=%ix%i", dpix, dpiy, wmm, hmm, w, h)
                     return dpix, dpiy
     return -1, -1
@@ -369,7 +382,7 @@ def get_number_of_desktops():
 def get_desktop_names():
     v = []
     if not is_Wayland():
-        v = ["Main"]
+        v = ("Main", )
         d = None
         try:
             d = _get_X11_root_property("_NET_DESKTOP_NAMES", "UTF8_STRING")
@@ -377,7 +390,7 @@ def get_desktop_names():
                 v = d.split(b"\0")
                 if len(v)>1 and v[-1]==b"":
                     v = v[:-1]
-                return [x.decode("utf8") for x in v]
+                return tuple(x.decode("utf8") for x in v)
         except Exception as e:
             screenlog.warn("failed to get desktop names: %s", e)
         screenlog("get_desktop_names() %s=%s", hexstr(d or ""), v)
@@ -407,7 +420,7 @@ def _get_xresources():
             from xpra.gtk_common.gtk_util import get_default_root_window
             root = get_default_root_window()
             value = prop_get(root, "RESOURCE_MANAGER", "latin1", ignore_errors=True)
-            log("RESOURCE_MANAGER=%s", value)
+            log(f"RESOURCE_MANAGER={value}")
             if value is None:
                 return None
             #parse the resources into a dict:
@@ -418,12 +431,12 @@ def _get_xresources():
                     continue
                 parts = option.split(":\t", 1)
                 if len(parts)!=2:
-                    log("skipped invalid option: '%s'", option)
+                    log(f"skipped invalid option: {option!r}")
                     continue
                 values[parts[0]] = parts[1]
             return values
         except Exception as e:
-            log("_get_xresources error: %s", e)
+            log(f"_get_xresources error: {e!r}")
     return None
 
 def get_cursor_size():
@@ -486,10 +499,30 @@ def system_bell(window, device, percent, _pitch, _duration, bell_class, bell_id,
         return False
 
 
+def pointer_grab(gdk_window):
+    if is_X11():
+        try:
+            from xpra.gtk_common.error import xsync
+            with xsync:
+                return X11WindowBindings().pointer_grab(gdk_window.get_xid())
+        except Exception:
+            log.error("Error: failed to grab pointer", exc_info=True)
+    return False
+
+def pointer_ungrab(_window):
+    if is_X11():
+        try:
+            from xpra.gtk_common.error import xsync
+            with xsync:
+                return X11WindowBindings().UngrabPointer()==0
+            return True
+        except Exception:
+            log.error("Error: failed to ungrab pointer", exc_info=True)
+    return False
+
+
 def _send_client_message(window, message_type, *values):
     try:
-        from xpra.x11.gtk_x11.gdk_display_source import init_gdk_display_source
-        init_gdk_display_source()
         from xpra.x11.bindings.window_bindings import constants #@UnresolvedImport
         X11Window = X11WindowBindings()
         root_xid = X11Window.getDefaultRootWindow()
@@ -563,6 +596,9 @@ def get_info():
     return i
 
 
+def suppress_event(*_args):
+    """ we'll use XI2 to receive events """
+
 class XI2_Window:
     def __init__(self, window):
         log("XI2_Window(%s)", window)
@@ -576,15 +612,12 @@ class XI2_Window:
         self.configured()
         #replace event handlers with XI2 version:
         self._do_motion_notify_event = window._do_motion_notify_event
-        window._do_motion_notify_event = self.noop
-        window._do_button_press_event = self.noop
-        window._do_button_release_event = self.noop
-        window._do_scroll_event = self.noop
+        window._do_motion_notify_event = suppress_event
+        window._do_button_press_event = suppress_event
+        window._do_button_release_event = suppress_event
+        window._do_scroll_event = suppress_event
         window.connect("destroy", self.cleanup)
 
-    def noop(self, *args):
-        pass
-
     def cleanup(self, *_args):
         for window in self.windows:
             self.XI2.disconnect(window)
@@ -651,8 +684,8 @@ class XI2_Window:
         client = window._client
         if client.readonly:
             return
-        pointer, relative_pointer, modifiers, buttons = window._pointer_modifiers(event)
-        wid = self.window.get_mouse_event_wid(*pointer)
+        pointer_data, modifiers, buttons = window._pointer_modifiers(event)
+        wid = self.window.get_mouse_event_wid(*pointer_data)
         #log("server_input_devices=%s, server_precise_wheel=%s",
         #    client.server_input_devices, client.server_precise_wheel)
         valuators = event.valuators
@@ -698,17 +731,14 @@ class XI2_Window:
         #send plain motion first, if any:
         if unused_valuators:
             xinputlog("do_xi_motion(%s, %s) wid=%s / focus=%s / window wid=%i, device=%s, pointer=%s, modifiers=%s, buttons=%s",
-                      event, device, wid, window._client._focused, window._id, event.device, pointer, modifiers, buttons)
-            pdata = pointer
-            if client.server_pointer_relative:
-                pdata = list(pointer)+list(relative_pointer)
-            packet = ["pointer-position", wid, pdata, modifiers, buttons] + self.get_pointer_extra_args(event)
+                      event, device, wid, window._client._focused, window._id, event.device, pointer_data, modifiers, buttons)
+            packet = ["pointer-position", wid, pointer_data, modifiers, buttons] + self.get_pointer_extra_args(event)
             client.send_mouse_position(packet)
         #now see if we have anything to send as a wheel event:
         if dx!=0 or dy!=0:
             xinputlog("do_xi_motion(%s, %s) wheel deltas: dx=%i, dy=%i", event, device, dx, dy)
             #normalize (xinput is always using 15 degrees?)
-            client.wheel_event(wid, dx/XINPUT_WHEEL_DIV, dy/XINPUT_WHEEL_DIV, event.device)
+            client.wheel_event(wid, dx/XINPUT_WHEEL_DIV, dy/XINPUT_WHEEL_DIV, pointer_data, event.device)
 
     def get_pointer_extra_args(self, event):
         def intscaled(f):
@@ -752,7 +782,7 @@ class ClientExtras:
         self.setup_dbus_signals()
 
     def ready(self):
-        pass
+        """ unused on xposix """
 
     def init_x11_filter(self):
         if self.x11_filter:
@@ -827,6 +857,15 @@ class ClientExtras:
             dbuslog.error(" %s", e)
             return
 
+        try:
+            import dbus
+            assert dbus
+        except ImportError as e:
+            dbuslog("setup_dbus_signals()", exc_info=True)
+            dbuslog.warn("Warning: cannot setup dbus signals")
+            dbuslog.warn(f" {e}")
+            return
+
         try:
             bus = init_system_bus()
             self.system_bus = bus
@@ -937,9 +976,9 @@ class ClientExtras:
             return False
         try:
             from xpra.gtk_common.error import xsync, XError
-            assert X11WindowBindings, "no X11 window bindings"
-            assert X11XI2Bindings, "no XI2 window bindings"
+            assert X11WindowBindings(), "no X11 window bindings"
             XI2 = X11XI2Bindings()
+            assert XI2, "no XI2 window bindings"
             #this may fail when windows are being destroyed,
             #ie: when another client disconnects because we are stealing the session
             try:
@@ -973,10 +1012,12 @@ class ClientExtras:
 
 
     def _get_xsettings(self):
-        try:
-            return self._xsettings_watcher.get_settings()
-        except:
-            log.error("failed to get XSETTINGS", exc_info=True)
+        xw = self._xsettings_watcher
+        if xw:
+            try:
+                return xw.get_settings()
+            except Exception:
+                log.error("failed to get XSETTINGS", exc_info=True)
         return None
 
     def _handle_xsettings_changed(self, *_args):
@@ -993,7 +1034,7 @@ class ClientExtras:
             value = prop_get(root, "RESOURCE_MANAGER", "latin1", ignore_errors=True)
             if value is not None:
                 return value.encode("utf-8")
-        except:
+        except (ImportError, UnicodeEncodeError):
             log.error("failed to get RESOURCE_MANAGER", exc_info=True)
         return None
 
