@@ -1,12 +1,9 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # This file is part of Xpra.
 # Copyright (C) 2014-2019 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-from xpra.scripts.config import python_platform
-#only imported to make sure we can get hold of a reference to the real "platform" module
-assert python_platform
 from xpra.version_util import get_version_info, get_platform_info, get_host_info
 from xpra.util import print_nested_dict
 
@@ -26,7 +23,7 @@ def main():
         print_nested_dict(pi)
         print("")
         print("Host:")
-        d = get_host_info()
+        d = get_host_info(2)
         #add os specific version info:
         from xpra.platform.info import get_version_info as pvinfo
         d.update(pvinfo())
@@ -34,5 +31,5 @@ def main():
     return 0
 
 
-if __name__ == "__main__":
+if __name__ == "__main__":  # pragma: no cover
     main()
