@@ -1,19 +1,18 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2012-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2012-2022 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
 
-from xpra.server.window.batch_config import DamageBatchConfig
+from xpra.server.window import batch_config
 from xpra.server.shadow.root_window_model import RootWindowModel
 from xpra.notifications.common import parse_image_path
 from xpra.platform.gui import get_native_notifier_classes, get_wm_name
 from xpra.platform.paths import get_icon_dir
 from xpra.server import server_features
-from xpra.os_util import PYTHON3
-from xpra.util import envint, envbool, DONE
+from xpra.util import envint, envbool, DONE, XPRA_STARTUP_NOTIFICATION_ID, XPRA_NEW_USER_NOTIFICATION_ID
 from xpra.log import Logger
 
 log = Logger("shadow")
@@ -21,11 +20,11 @@ notifylog = Logger("notify")
 mouselog = Logger("mouse")
 cursorlog = Logger("cursor")
 
-REFRESH_DELAY = envint("XPRA_SHADOW_REFRESH_DELAY", 50)
 NATIVE_NOTIFIER = envbool("XPRA_NATIVE_NOTIFIER", True)
 POLL_POINTER = envint("XPRA_POLL_POINTER", 20)
 CURSORS = envbool("XPRA_CURSORS", True)
 SAVE_CURSORS = envbool("XPRA_SAVE_CURSORS", False)
+NOTIFY_STARTUP = envbool("XPRA_SHADOW_NOTIFY_STARTUP", True)
 
 
 SHADOWSERVER_BASE_CLASS = object
@@ -36,22 +35,25 @@ if server_features.rfb:
 
 class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
 
+    #20 fps unless the client specifies more:
+    DEFAULT_REFRESH_RATE = 20
+
     def __init__(self, root_window, capture=None):
-        SHADOWSERVER_BASE_CLASS.__init__(self)
+        super().__init__()
         self.capture = capture
         self.root = root_window
+        self.window_matches = None
         self.mapped = []
         self.pulseaudio = False
         self.sharing = True
-        self.refresh_delay = REFRESH_DELAY
+        self.refresh_delay = 1000//self.DEFAULT_REFRESH_RATE
         self.refresh_timer = None
         self.notifications = False
         self.notifier = None
         self.pointer_last_position = None
         self.pointer_poll_timer = None
         self.last_cursor_data = None
-        DamageBatchConfig.ALWAYS = True             #always batch
-        DamageBatchConfig.MIN_DELAY = 50            #never lower than 50ms
+        batch_config.ALWAYS = True             #always batch
 
     def init(self, opts):
         if SHADOWSERVER_BASE_CLASS!=object:
@@ -60,11 +62,18 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
         self.notifications = bool(opts.notifications)
         if self.notifications:
             self.make_notifier()
+        log("init(..) session_name=%s", opts.session_name)
         if opts.session_name:
             self.session_name = opts.session_name
         else:
             self.guess_session_name()
 
+    def run(self):
+        if NOTIFY_STARTUP:
+            from gi.repository import GLib
+            GLib.timeout_add(1000, self.notify_startup_complete)
+        return super().run()
+
     def cleanup(self):
         for wid in self.mapped:
             self.stop_refresh(wid)
@@ -78,13 +87,13 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
             capture.clean()
 
 
-    def guess_session_name(self, _procs=None):
-        self.session_name = get_wm_name()
+    def guess_session_name(self, procs=None):
+        log("guess_session_name(%s)", procs)
+        self.session_name = get_wm_name()       # pylint: disable=assignment-from-none
+        log("get_wm_name()=%s", self.session_name)
 
     def get_server_mode(self):
-        if PYTHON3:
-            return "GTK3 shadow"
-        return "GTK2 shadow"
+        return "shadow"
 
     def print_screen_info(self):
         if not server_features.display:
@@ -98,6 +107,8 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
             log.info(" on display '%s' of size %ix%i", display, w, h)
         else:
             log.info(" on display of size %ix%i", w, h)
+        if self.window_matches:
+            return
         try:
             l = len(self._id_to_window)
         except AttributeError as e:
@@ -110,13 +121,23 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
                 x, y, w, h = window.geometry
                 log.info("  %-16s %4ix%-4i at %4i,%-4i", title, w, h, x, y)
 
+    def apply_refresh_rate(self, ss):
+        rrate = super().apply_refresh_rate(ss)
+        if rrate>0:
+            #adjust refresh delay to try to match:
+            self.set_refresh_delay(max(10, 1000//rrate))
+
+
     def make_hello(self, _source):
         return {"shadow" : True}
 
     def get_info(self, _proto=None):
         info = {
-            "notifications" : self.notifications,
+            "sharing"       : self.sharing is not False,
+            "refresh-delay" : self.refresh_delay,
             }
+        if self.pointer_last_position:
+            info["pointer-last-position"] = self.pointer_last_position
         return info
 
 
@@ -124,8 +145,13 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
         #we export the whole desktop as a window:
         return 0, 0
 
-    def watch_keymap_changes(self):
-        pass
+    def _keys_changed(self):
+        from xpra.server.mixins.input_server import InputServer
+        if isinstance(self, InputServer):
+            InputServer._keys_changed(self)
+            from xpra.platform.keyboard import Keyboard
+            log.info("the keymap has been changed: %s", Keyboard().get_layout_spec()[0])
+
 
     def timeout_add(self, *args):
         #usually done via gobject
@@ -147,6 +173,8 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
     def notify_setup_error(self, exception):
         notifylog("notify_setup_error(%s)", exception)
         notifylog.info("notification forwarding is not available")
+        if str(exception).endswith("is already claimed on the session bus"):
+            log.info(" the interface is already claimed")
 
     def make_notifier(self):
         nc = self.get_notifier_classes()
@@ -156,7 +184,7 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
                 self.notifier = x()
                 notifylog("notifier=%s", self.notifier)
                 break
-            except:
+            except Exception:
                 notifylog("failed to instantiate %s", x, exc_info=True)
 
     def get_notifier_classes(self):
@@ -172,8 +200,8 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
         #directly on the screen we shadow
         notifylog("notify_new_user(%s) notifier=%s", ss, self.notifier)
         if self.notifier:
-            tray = self.get_notification_tray()
-            nid = 0
+            tray = self.get_notification_tray()     #pylint: disable=assignment-from-none
+            nid = XPRA_NEW_USER_NOTIFICATION_ID
             title = "User '%s' connected to the session" % (ss.name or ss.username or ss.uuid)
             body = "\n".join(ss.get_connect_info())
             actions = []
@@ -187,6 +215,25 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
     def get_notification_tray(self):
         return None
 
+    def notify_startup_complete(self):
+        self.do_notify_startup("Xpra shadow server is ready", replaces_nid=XPRA_STARTUP_NOTIFICATION_ID)
+
+    def do_notify_startup(self, title, body="", replaces_nid=0):
+        #overriden here so we can show the notification
+        #directly on the screen we shadow
+        notifylog("do_notify_startup%s", (title, body, replaces_nid))
+        if self.notifier:
+            tray = self.get_notification_tray()     #pylint: disable=assignment-from-none
+            nid = XPRA_STARTUP_NOTIFICATION_ID
+            actions = []
+            hints = {}
+            icon = None
+            icon_filename = os.path.join(get_icon_dir(), "server-connected.png")
+            if os.path.exists(icon_filename):
+                icon = parse_image_path(icon_filename)
+            self.notifier.show_notify("", tray, nid, "Xpra", replaces_nid, "",
+                                      title, body, actions, hints, 10*1000, icon)
+
 
     ############################################################################
     # refresh
@@ -195,9 +242,12 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
         log("start_refresh(%i) mapped=%s, timer=%s", wid, self.mapped, self.refresh_timer)
         if wid not in self.mapped:
             self.mapped.append(wid)
+        self.start_refresh_timer()
+        self.start_poll_pointer()
+
+    def start_refresh_timer(self):
         if not self.refresh_timer:
             self.refresh_timer = self.timeout_add(self.refresh_delay, self.refresh)
-        self.start_poll_pointer()
 
     def set_refresh_delay(self, v):
         assert 0<v<10000
@@ -215,8 +265,11 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
         except KeyError:
             pass
         if not self.mapped:
-            self.cancel_refresh_timer()
-            self.cancel_poll_pointer()
+            self.no_windows()
+
+    def no_windows(self):
+        self.cancel_refresh_timer()
+        self.cancel_poll_pointer()
 
     def cancel_refresh_timer(self):
         t = self.refresh_timer
@@ -275,7 +328,9 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
             if rwm:
                 mouselog("poll_pointer_position() wid=%i, position=%s, relative=%s", wid, (x, y), (rx, ry))
                 for ss in self._server_sources.values():
-                    ss.update_mouse(wid, x, y, rx, ry)
+                    um = getattr(ss, "update_mouse", None)
+                    if um:
+                        um(wid, x, y, rx, ry)
             else:
                 mouselog("poll_pointer_position() model not found for position=%s", (x, y))
         else:
@@ -284,7 +339,7 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
 
     def poll_cursor(self):
         prev = self.last_cursor_data
-        curr = self.do_get_cursor_data()
+        curr = self.do_get_cursor_data()        #pylint: disable=assignment-from-none
         self.last_cursor_data = curr
         def cmpv(lcd):
             if not lcd:
@@ -297,8 +352,8 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
             fields = ("x", "y", "width", "height", "xhot", "yhot", "serial", "pixels", "name")
             if len(prev or [])==len(curr or []) and len(prev or [])==len(fields):
                 diff = []
-                for i in range(len(prev)):
-                    if prev[i]!=curr[i]:
+                for i, prev_value in enumerate(prev):
+                    if prev_value!=curr[i]:
                         diff.append(fields[i])
                 cursorlog("poll_cursor() attributes changed: %s", diff)
             if SAVE_CURSORS and curr:
@@ -312,7 +367,7 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
                     from PIL import Image
                     img = Image.frombuffer("RGBA", (w, h), pixels, "raw", "BGRA", 0, 1)
                     img.save("cursor-%#x.png" % serial, format="PNG")
-            for ss in self._server_sources.values():
+            for ss in self.window_sources():
                 ss.send_cursor()
 
     def do_get_cursor_data(self):
@@ -333,7 +388,8 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
                 log.warn("Warning: shadowing your own display can be quite confusing")
                 clipboard = self._clipboard_helper and c.boolget("clipboard", True)
                 if clipboard:
-                    log.warn("clipboard sharing cannot be enabled! (consider using the --no-clipboard option)")
+                    log.warn("Warning: clipboard sharing cannot be enabled!")
+                    log.warn(" consider using the --no-clipboard option")
                     c["clipboard"] = False
             else:
                 log.warn("This client is running within the Xpra server %s", server_uuid)
@@ -342,20 +398,21 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
     def parse_screen_info(self, ss):
         try:
             log.info(" client root window size is %sx%s", *ss.desktop_size)
-        except:
+        except Exception:
             log.info(" unknown client desktop size")
+        self.apply_refresh_rate(ss)
         return self.get_root_window_size()
 
     def _process_desktop_size(self, proto, packet):
         #just record the screen size info in the source
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss and len(packet)>=4:
             ss.set_screen_sizes(packet[3])
 
 
     def set_keyboard_repeat(self, key_repeat):
         """ don't override the existing desktop """
-        pass
+        pass    #pylint: disable=unnecessary-pass
 
     def set_keymap(self, server_source, force=False):
         log("set_keymap%s", (server_source, force))
@@ -384,15 +441,22 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
 
     def _add_new_window(self, window):
         self._add_new_window_common(window)
-        self._send_new_window_packet(window)
+        if window.get("override-redirect", False):
+            self._send_new_or_window_packet(window)
+        else:
+            self._send_new_window_packet(window)
 
     def _send_new_window_packet(self, window):
         geometry = window.get_geometry()
         self._do_send_new_window_packet("new-window", window, geometry)
 
+    def _send_new_or_window_packet(self, window):
+        geometry = window.get_property("geometry")
+        self._do_send_new_window_packet("new-override-redirect", window, geometry)
+
     def _process_window_common(self, wid):
         window = self._id_to_window.get(wid)
-        assert window is not None
+        assert window is not None, "wid %s does not exist" % wid
         return window
 
     def _process_map_window(self, proto, packet):
