@@ -1,11 +1,10 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2021 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-from xpra.os_util import strtobytes
-from xpra.util import get_screen_info, envint, first_time, iround, typedict
+from xpra.util import get_screen_info, envint, first_time, typedict, net_utf8
 from xpra.server.source.stub_source_mixin import StubSourceMixin
 from xpra.log import Logger
 
@@ -30,6 +29,7 @@ class ClientDisplayMixin(StubSourceMixin):
         self.desktop_size_unscaled = None
         self.desktop_size_server = None
         self.screen_sizes = ()
+        self.monitors = {}
         self.screen_resize_bigger = True
         self.desktops = 1
         self.desktop_names = ()
@@ -44,6 +44,7 @@ class ClientDisplayMixin(StubSourceMixin):
             "desktop_names" : self.desktop_names,
             "randr_notify"  : self.randr_notify,
             "opengl"        : self.opengl_props,
+            "monitors"      : self.monitors,
             }
         info.update(get_screen_info(self.screen_sizes))
         if self.desktop_mode_size:
@@ -65,12 +66,34 @@ class ClientDisplayMixin(StubSourceMixin):
         self.desktop_size_unscaled = c.intpair("desktop_size.unscaled")
         self.screen_resize_bigger = c.boolget("screen-resize-bigger", True)
         self.set_screen_sizes(c.tupleget("screen_sizes"))
-        self.set_desktops(c.intget("desktops", 1), c.strtupleget("desktop.names"))
+        self.set_monitors(c.dictget("monitors"))
+        desktop_names = tuple(net_utf8(x) for x in c.tupleget("desktop.names"))
+        self.set_desktops(c.intget("desktops", 1), desktop_names)
         self.show_desktop_allowed = c.boolget("show-desktop")
         self.icc = c.dictget("icc", {})
         self.display_icc = c.dictget("display-icc", {})
         self.opengl_props = c.dictget("opengl", {})
 
+    def set_monitors(self, monitors):
+        self.monitors = {}
+        if monitors:
+            for i, mon_def in monitors.items():
+                vdef = self.monitors.setdefault(i, {})
+                td = typedict(mon_def)
+                for attr, conv in {
+                    "geometry"  : td.inttupleget,
+                    "primary"   : td.boolget,
+                    "refresh-rate"  : td.intget,
+                    "scale-factor"  : td.intget,
+                    "width-mm"      : td.intget,
+                    "height-mm"     : td.intget,
+                    "manufacturer"  : td.strget,
+                    "model"         : td.strget,
+                    "subpixel-layout" : td.strget,
+                    "workarea"      : td.inttupleget,
+                    }.items():
+                    vdef[attr] = conv(attr)
+        log("set_monitors(%s) monitors=%s", monitors, self.monitors)
 
     def set_screen_sizes(self, screen_sizes):
         log("set_screen_sizes(%s)", screen_sizes)
@@ -82,7 +105,7 @@ class ClientDisplayMixin(StubSourceMixin):
         def dpi(size_pixels, size_mm):
             if size_mm==0:
                 return 0
-            return int(size_pixels * 254 / size_mm / 10)
+            return round(size_pixels * 25.4 / size_mm)
         for i,screen in enumerate(list(screen_sizes)):
             if len(screen)<10:
                 continue
@@ -95,17 +118,20 @@ class ClientDisplayMixin(StubSourceMixin):
                     log.warn("Warning: ignoring invalid screen size %ix%i mm", wmm, hmm)
                 if monitors:
                     #[plug_name, xs(geom.x), ys(geom.y), xs(geom.width), ys(geom.height), wmm, hmm]
-                    wmm = sum(monitor[5] for monitor in monitors)
-                    hmm = sum(monitor[6] for monitor in monitors)
+                    wmm = round(sum(monitor[5] for monitor in monitors))
+                    hmm = round(sum(monitor[6] for monitor in monitors))
                     xdpi = dpi(sw, wmm)
                     ydpi = dpi(sh, hmm)
                 if xdpi<MIN_DPI or xdpi>MAX_DPI or ydpi<MIN_DPI or ydpi>MAX_DPI:
                     #still invalid, generate one from DPI=96
-                    wmm = iround(sw*25.4/96)
-                    hmm = iround(sh*25.4/96)
+                    wmm = round(sw*25.4/96)
+                    hmm = round(sh*25.4/96)
                 if warn:
                     log.warn(" using %ix%i mm", wmm, hmm)
                 screen = list(screen)
+                #make sure values are integers:
+                for j in range(4):
+                    screen[j] = round(screen[j])
                 screen[3] = wmm
                 screen[4] = hmm
                 self.screen_sizes[i] = tuple(screen)
@@ -113,17 +139,7 @@ class ClientDisplayMixin(StubSourceMixin):
 
     def set_desktops(self, desktops, desktop_names):
         self.desktops = desktops or 1
-        #older clients send strings,
-        #newer clients send bytes...
-        def b(v):
-            try :
-                return strtobytes(v).decode("utf8")
-            except UnicodeDecodeError:
-                return v
-        if desktop_names:
-            self.desktop_names = [b(d) for d in desktop_names]
-        else:
-            self.desktop_names = []
+        self.desktop_names = tuple(net_utf8(d) for d in (desktop_names or ()))
 
     def updated_desktop_size(self, root_w, root_h, max_w, max_h):
         log("updated_desktop_size%s randr_notify=%s, desktop_size=%s",
