@@ -1,6 +1,6 @@
 # This file is part of Xpra.
 # Copyright (C) 2008, 2009 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2011-2021 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2011-2022 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -10,11 +10,12 @@ from socket import gethostname
 from gi.repository import GObject, Gdk, GLib
 
 from xpra.util import envbool, first_time
-from xpra.os_util import bytestostr
+from xpra.os_util import bytestostr, get_proc_cmdline
 from xpra.x11.common import Unmanageable
 from xpra.gtk_common.gobject_util import one_arg_signal, two_arg_signal
 from xpra.gtk_common.error import XError, xsync, xswallow
 from xpra.x11.bindings.window_bindings import X11WindowBindings, constants, SHAPE_KIND #@UnresolvedImport
+from xpra.x11.bindings.res_bindings import ResBindings #@UnresolvedImport
 from xpra.x11.models.model_stub import WindowModelStub
 from xpra.x11.gtk_x11.composite import CompositeHelper
 from xpra.x11.gtk_x11.prop import prop_get, prop_set, prop_type_get, PYTHON_TYPES
@@ -34,9 +35,21 @@ X11Window = X11WindowBindings()
 em = Gdk.EventMask
 ADDMASK = em.STRUCTURE_MASK | em.PROPERTY_CHANGE_MASK | em.FOCUS_CHANGE_MASK | em.POINTER_MOTION_MASK
 
+XRes = ResBindings()
+if not XRes.check_xres():
+    log.warn("Warning: X Resource Extension missing or too old")
+    XRes = None
+
+try:
+    from xpra.platform.xposix.proc import get_parent_pid
+except ImportError:
+    log("proc.get_parent_pid is not available", exc_info=True)
+    get_parent_pid = None
+
 FORCE_QUIT = envbool("XPRA_FORCE_QUIT", True)
 XSHAPE = envbool("XPRA_XSHAPE", True)
 FRAME_EXTENTS = envbool("XPRA_FRAME_EXTENTS", True)
+OPAQUE_REGION = envbool("XPRA_OPAQUE_REGION", True)
 
 # Re-stacking:
 Above = 0
@@ -76,7 +89,7 @@ grablog("detail constants: %s", DETAIL_CONSTANTS)
 
 #these properties are not handled, and we don't want to spam the log file
 #whenever an app decides to change them:
-PROPERTIES_IGNORED = os.environ.get("XPRA_X11_PROPERTIES_IGNORED", "_NET_WM_OPAQUE_REGION").split(",")
+PROPERTIES_IGNORED = [x for x in os.environ.get("XPRA_X11_PROPERTIES_IGNORED", "").split(",") if x]
 #make it easier to debug property changes, just add them here:
 #ie: {"WM_PROTOCOLS" : ["atom"]}
 X11_PROPERTIES_DEBUG = {}
@@ -128,11 +141,20 @@ class CoreX11WindowModel(WindowModelStub):
         "client-machine": (GObject.TYPE_PYOBJECT,
                 "Host where client process is running", "",
                 GObject.ParamFlags.READABLE),
-        #from _NET_WM_PID
+        #from XResGetClientPid
         "pid": (GObject.TYPE_INT,
                 "PID of owning process", "",
                 -1, 65535, -1,
                 GObject.ParamFlags.READABLE),
+        "ppid": (GObject.TYPE_INT,
+                "PID of parent process", "",
+                -1, 65535, -1,
+                GObject.ParamFlags.READABLE),
+        #from _NET_WM_PID
+        "wm-pid": (GObject.TYPE_INT,
+                "PID of owning process", "",
+                -1, 65535, -1,
+                GObject.ParamFlags.READABLE),
         #from _NET_WM_NAME or WM_NAME
         "title": (GObject.TYPE_PYOBJECT,
                 "Window title (unicode or None)", "",
@@ -165,6 +187,10 @@ class CoreX11WindowModel(WindowModelStub):
         "allowed-actions": (GObject.TYPE_PYOBJECT,
                 "Supported WM actions", "",
                 GObject.ParamFlags.READWRITE),
+        #synced to "_NET_WM_OPAQUE_REGION"
+        "opaque-region": (GObject.TYPE_PYOBJECT,
+                "Compositor can assume that there is no transparency for this region", "",
+                GObject.ParamFlags.READWRITE),
            }
 
     __common_signals__ = {
@@ -193,19 +219,23 @@ class CoreX11WindowModel(WindowModelStub):
     #things that we expose:
     _property_names         = [
         "xid", "depth", "has-alpha",
-        "client-machine", "pid",
+        "client-machine", "pid", "ppid", "wm-pid",
         "title", "role",
         "command", "shape",
         "class-instance", "protocols",
+        "opaque-region",
         ]
     #exposed and changing (should be watched for notify signals):
-    _dynamic_property_names = ["title", "command", "shape", "class-instance", "protocols"]
+    _dynamic_property_names = ["title", "command", "shape", "class-instance", "protocols", "opaque-region"]
     #should not be exported to the clients:
     _internal_property_names = ["frame", "allowed-actions"]
     _initial_x11_properties = ["_NET_WM_PID", "WM_CLIENT_MACHINE",
                                #_NET_WM_NAME is redundant, as it calls the same handler as "WM_NAME"
                                "WM_NAME", "_NET_WM_NAME",
-                               "WM_PROTOCOLS", "WM_CLASS", "WM_WINDOW_ROLE"]
+                               "WM_PROTOCOLS", "WM_CLASS", "WM_WINDOW_ROLE",
+                               "_NET_WM_OPAQUE_REGION",
+                               "WM_COMMAND",
+                               ]
     _DEFAULT_NET_WM_ALLOWED_ACTIONS = []
     _MODELTYPE = "Core"
     _scrub_x11_properties       = [
@@ -371,9 +401,13 @@ class CoreX11WindowModel(WindowModelStub):
         metalog("read_initial_X11_properties() core")
         #immutable ones:
         depth = X11Window.get_depth(self.xid)
-        metalog("initial X11 properties: xid=%#x, depth=%i", self.xid, depth)
+        pid = XRes.get_pid(self.xid) if XRes else -1
+        ppid = get_parent_pid(pid) if pid and get_parent_pid else 0
+        metalog("initial X11 properties: xid=%#x, depth=%i, pid=%i, ppid=%i", self.xid, depth, pid, ppid)
         self._updateprop("depth", depth)
         self._updateprop("xid", self.xid)
+        self._updateprop("pid", pid)
+        self._updateprop("ppid", ppid)
         self._updateprop("has-alpha", depth==32)
         self._updateprop("allowed-actions", self._DEFAULT_NET_WM_ALLOWED_ACTIONS)
         self._updateprop("shape", self._read_xshape())
@@ -552,7 +586,7 @@ class CoreX11WindowModel(WindowModelStub):
     def _handle_pid_change(self):
         pid = self.prop_get("_NET_WM_PID", "u32") or -1
         metalog("_NET_WM_PID=%s", pid)
-        self._updateprop("pid", pid)
+        self._updateprop("wm-pid", pid)
 
     def _handle_client_machine_change(self):
         client_machine = self.prop_get("WM_CLIENT_MACHINE", "latin1")
@@ -584,6 +618,9 @@ class CoreX11WindowModel(WindowModelStub):
         metalog("WM_COMMAND=%s", command)
         if command:
             command = command.strip("\0")
+        else:
+            pid = self.get_property("pid")
+            command = b" ".join(get_proc_cmdline(pid) or ())
         self._updateprop("command", command)
 
     def _handle_class_change(self):
@@ -593,6 +630,16 @@ class CoreX11WindowModel(WindowModelStub):
         metalog("WM_CLASS=%s", class_instance)
         self._updateprop("class-instance", class_instance)
 
+    def _handle_opaque_region_change(self):
+        rectangles = []
+        v = tuple(self.prop_get("_NET_WM_OPAQUE_REGION", ["u32"]) or [])
+        if OPAQUE_REGION and len(v)%4==0:
+            while v:
+                rectangles.append(v[:4])
+                v = v[4:]
+        metalog("_NET_WM_OPAQUE_REGION(%s)=%s (OPAQUE_REGION=%s)", v, rectangles, OPAQUE_REGION)
+        self._updateprop("opaque-region", tuple(rectangles))
+
     #these handlers must not generate X11 errors (must use XSync)
     _x11_property_handlers = {
         "_NET_WM_PID"       : _handle_pid_change,
@@ -603,6 +650,7 @@ class CoreX11WindowModel(WindowModelStub):
         "WM_PROTOCOLS"      : _handle_protocols_change,
         "WM_COMMAND"        : _handle_command_change,
         "WM_CLASS"          : _handle_class_change,
+        "_NET_WM_OPAQUE_REGION" : _handle_opaque_region_change,
         }
 
 
@@ -753,14 +801,13 @@ class CoreX11WindowModel(WindowModelStub):
 
     def request_close(self):
         if "WM_DELETE_WINDOW" in self.get_property("protocols"):
-            with xswallow:
-                send_wm_delete_window(self.client_window)
+            self.send_delete()
         else:
             title = self.get_property("title")
             xid = self.get_property("xid")
             if FORCE_QUIT:
                 log.info("window %#x ('%s') does not support WM_DELETE_WINDOW", xid, title)
-                log.info(" using force_quit")
+                log.info(" using force quit")
                 # You don't wanna play ball?  Then no more Mr. Nice Guy!
                 self.force_quit()
             else:
@@ -768,18 +815,33 @@ class CoreX11WindowModel(WindowModelStub):
                 log.warn(" it does not support WM_DELETE_WINDOW")
                 log.warn(" and FORCE_QUIT is disabled")
 
+    def send_delete(self):
+        with xswallow:
+            send_wm_delete_window(self.client_window)
+
+    def XKill(self):
+        with xswallow:
+            X11Window.XKillClient(self.xid)
+
     def force_quit(self):
-        pid = self.get_property("pid")
         machine = self.get_property("client-machine")
+        pid = self.get_property("pid")
+        if pid<=0:
+            #we could fallback to _NET_WM_PID
+            #but that would be unsafe
+            log.warn("Warning: cannot terminate window %#x, no pid found", self.xid)
+            if machine:
+                log.warn(" WM_CLIENT_MACHINE=%s", machine)
+            pid = self.get_property("wm-pid")
+            if pid>0:
+                log.warn(" _NET_WM_PID=%s", pid)
+            return
+        if pid==os.getpid():
+            log.warn("Warning: force_quit is refusing to kill ourselves!")
+            return
         localhost = gethostname()
         log("force_quit() pid=%s, machine=%s, localhost=%s", pid, machine, localhost)
-        def XKill():
-            with xswallow:
-                X11Window.XKillClient(self.xid)
-        if pid > 0 and machine is not None and machine == localhost:
-            if pid==os.getpid():
-                log.warn("Warning: force_quit is refusing to kill ourselves!")
-                return
+        if machine is not None and machine == localhost:
             if self._kill_count==0:
                 #first time around: just send a SIGINT and hope for the best
                 try:
@@ -794,7 +856,7 @@ class CoreX11WindowModel(WindowModelStub):
                 except OSError as e:
                     log.warn("Warning: failed to kill(SIGKILL) client with pid %s", pid)
                     log.warn(" %s", e)
-                XKill()
+                self.XKill()
             self._kill_count += 1
             return
-        XKill()
+        self.XKill()
